/**************************************************************************
 *									  *
 * 		 Copyright (C) 1989, Silicon Graphics, Inc.		  *
 *									  *
 *  These coded instructions, statements, and computer programs  contain  *
 *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
 *  are protected by Federal copyright law.  They  may  not be disclosed  *
 *  to  third  parties  or copied or duplicated in any form, in whole or  *
 *  in part, without the prior written consent of Silicon Graphics, Inc.  *
 *									  *
 **************************************************************************/
/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/
#ifndef __SYS_PROC_H__
#define __SYS_PROC_H__

#ident	"$Revision: 3.114 $"

/*	One structure allocated per active process. It contains all
**	data needed about the process while the process may be swapped
**	out.  Other per process data (user.h) may swapped with the
**	process.
*/

/*
 * Proc structure and maintenance variables description.
 *
 * The fields in a proc entry are protected by an interwoven set of
 * locks and semaphores.  Separate locks and semaphores are used to
 * squeeze decent performance out of proc table and entry handling.
 * The locks and semaphores, and what they protect, are:
 *
 * The following locks are used solely within a single proc entry,
 * and are thus contained within that entry:
 *
 * p_siglck - This lock is used during signal handling.  It protects
 *	      the fields p_sig, p_sigmask, p_hold, p_sigign, p_poll,
 *	      p_sigcatch, and p_nexit.  The lock must be held before
 *	      examining or modifying any of these variables at all times.
 *	      This lock also is used to protect the p_stat, p_flag and p_flag2
 *	      fields where protection is necessary.
 *	      Protects p_sonproc field which marks which processors the
 *	      given proc entry has threads running on.
 *
 * p_sema   - This semaphore protects the parent-child-sibling chain
 *	      rooted at the current proc entry.
 *	      Thus, in order to access ones parent, the parent p_sema
 *	      must be obtained - though while doing so the parent may have
 *	      exited.
 *
 *	      It is important
 *	      to realize that some fields in the proc structure can only
 *	      be modified by the process itself, and thus there is no
 *            protection needed, and the semaphore should not be acquired.
 *
 * p_parlck - This spinlock protects the p_parent field in the current
 *	      proc entry.
 *
 * p_wait   - This semaphore is used exclusively for the wait(2) system call
 *
 * p_aspacelock -
 *	grab the lock for MR_UPDATE, if the operation will change:
 *		1) pregion list
 *		2) regva, pgoff, pgsz, r_list
 *	Otherwise grab it MR_ACCESS if accessing any translations
 *
 * Links:
 *	      Free proc structures are threaded on p_link (== p_flink)
 *	      Runnable procs are threaded on p_link
 *	      Sleeping procs are double threaded on p_flink & p_blink
 *	      Active procs are threaded on p_active
 *	      Exiting procs are threaded on p_link
 *
 * The following locks are used in global proc table management:
 *
 * pfreelck - This lock protects the free process list.  This is a list
 *            of all free process slots, to avoid an unnecessary search
 *	      when creating a new process.  The free process list is
 *	      strung through the p_active field of the proc structure.
 *
 * pactlck  - This lock protects the active process list.  This is a list
 *	      of all active processes, and is primarly used when creating
 *	      a new process to reduce the search time for a new PID.  It
 *	      is strung through the p_active field of the proc structure.
 *
 * mpidlock - This lock protects the PID variable 'mpid' which holds the
 *	      next valid PID.
 */

#include "sys/types.h"
#include "sys/param.h"
#include "sys/sema.h"
#include "sys/time.h"
#include "sys/immu.h"
#include "sys/ssleep.h"

typedef struct	proc {
	char	p_stat;			/* status of process.	     */
	unchar	p_pri;			/* priority, lower # is high */
	char	p_cpu;			/* cpu usage for scheduling */
	char	p_nice;			/* nice for cpu usage */
	uint	p_flag;			/* flags defined below.	     */
	uid_t	p_uid;			/* real user id */
	uid_t	p_suid;			/* saved (effective) uid from exec */
        short   p_pgrp;                 /* name of process group leader */
	short	p_pid;			/* unique process id */
	short	p_ppid;			/* process id of parent*/
	gid_t	p_sgid;			/* saved (effective) gid from exec */
	struct	proc	*p_flink;	/* linked list of sleeping processes */
	struct	proc	*p_blink;	/* linked list of sleeping processes */
	caddr_t	p_wchan;		/* Wait addr for sleeping   */
					/* processes.		    */
	caddr_t	p_w2chan;		/* Wait addr for sleeping   */
					/* processes. (sleep hash)  */
	lock_t	p_parlck;		/* lock to protect p_parent */
	struct	proc	*p_parent;	/* ptr to parent process */
	struct	proc	*p_child;	/* ptr to first child process */
	struct	proc	*p_sibling;	/* ptr to next sibling proc on chain */
	struct	proc	*p_active;	/* active process chain	    */

	int	p_clktim;		/* time to alarm clock signal */
	uint	p_size;			/* size of swappable image in pages */
	uint	p_rsssize;		/* resident size of swappable image  */
					/* in pages.		    */
	uint		p_maxrsssize;	/* max rss - from u_limit	*/
	unsigned char	*p_tlbpid;	/* pointer to tlbpid array	*/
	pde_t		p_upgs[USIZE];	/* pdes for upage/user kstk	*/
	signed char	p_resident;	/* process cannot be swapped	*/
	unchar		p_upri;		/* user-level priority */
	cpuid_t		p_sonproc;	/* processor process is running on */
	char		p_vmsched;	/* sched info from VM system	*/
	struct pregion	*p_region;	/* process regions		*/
	struct pregion	*p_lregion;	/* last pregion accessed	*/
	sema_t		p_preglock;	/* pregion lock			*/
	int		p_nlockpg;	/* # locked pages		*/
	struct pregion	*p_stkreg;	/* processes's stack region	*/
	short		p_epid;		/* effective pid; normally same as
					 * p_pid; for servers, the system that
					 * sent the msg
					 */
	sysid_t		p_sysid;	/* normally - same as sysid */
					/* if server - system that sent msg */
	char		p_time;		/* resident time for scheduling */
	cpuid_t		p_mustrun;	/* cpu to run on, set by user/kernel */
	unchar		p_onrq;		/* set if on runq, runq.lock protects */
	cpuid_t		p_lastrun;	/* last processor flag */
	char		p_prtn;		/* semaphore return value	*/
	char		p_sqself;	/* If I'm using my stack */
	unchar		p_rtpri;	/* real-time priority */
	signed char	p_nexit;	/* delay exit */
	lock_t		p_siglck;	/* signal handling lock		*/
	ulong		p_sigmask;	/* tracing signal mask for /proc */
	ulong		p_hold;		/* hold signal bit mask		*/
	ulong		p_sig;		/* signals pending to this	*/
					/* process			*/
	ulong		p_sigign;	/* signals to be ignored	*/
	ulong		p_sigcatch;	/* signals to be caught		*/

	sema_t		p_sema;		/* semaphore on entry		*/
	sema_t		p_wait;		/* waiting for something to happen */
	ushort		p_whystop;	/* Reason for process stop */
	ushort		p_whatstop;	/* More detailed reason */
	union {
		struct rfs_server {
			struct	rcvd  *minwd;
					/* server msg arrived on this queue */
			struct	proc  *rlink;
					/* linked list for server */
		} rs;
		struct exit_info {
			/*
			 * utime and stime hold the same information as 
			 * the rusage structure and are present for the ps(1)
			 * command.
			 */
			time_t	utime;
					/* user time, this proc */
			time_t	stime;
					/* system time, this proc */
			short	xstat;
					/* exit status for wait */
			struct rusage *ru_info;
		} ei;
		struct itimer_info {
			struct	itimerval realtimer;
					/* time to next real alarm */
			lock_t	itimerlock;
					/* ITIMER_REAL handling lock */
			int	interval_tick;
			int	interval_unit;
		} ii;
	} ps;

	struct ssleep_s	*p_block;	/* block struct for blockproc() */
	struct proc	*p_slink;	/* share group link */
	uint		p_shmask;	/* share group share mask */
	struct shaddr_s	*p_shaddr;	/* pointer to shared group desc */

	ushort		p_tslice;	/* process time slice size */
	char		p_fp;		/* generate SIGFPE on fp interrupts */
	struct gfx_gfx *p_gfx;		/* head of graphics board list for this
					   proc */
	void		*p_aiop;	/* reserved for asyncio */
	int		p_aiocnt;	/* reserved for asyncio */
	caddr_t		*p_segtbl;	/* working segment table */
	uint		p_dvec;		/* scheduling discipline vector */
	struct prda	*p_prda;	/* K0 addr of locked-down prda */
	ushort		p_segtype;	/* segment table type */
	ushort		p_jwhatstop;	/* job control stop signal */
	uint		p_flag2;	/* more flags defined below.	     */
	struct pwatch_s	*p_watch;	/* watchpoint list */
	struct proc 	*p_pgflink;	/* process group chain */
	struct proc 	*p_pgblink;	/* process group chain */
	struct pgroup_s *p_pgroup;	/* process group structure */
	struct proc	*p_poll;	/* process select/polling via /proc */
	mrlock_t	*p_aspacelock;	/* address space lock */
	mrlock_t	p_alock;	/* non shared group aspace lock */
	struct session_s *p_session;	/* session */
	struct user 	*p_user;	/* K2SEG address for u area (R4K only)*/
} proc_t;

#if defined(R4000)
#define proc_to_user(p) ((p)->p_user)
#endif
#if defined(R3000)
#define proc_to_user(p) ((user_t *)pdetokv(&(p)->p_upgs[0]))
#endif

#define	p_link		p_flink
#define p_minwd		ps.rs.minwd
#define p_rlink		ps.rs.rlink
#define p_utime		ps.ei.utime
#define p_stime		ps.ei.stime
#define p_ru		ps.ei.ru_info
#define p_xstat		ps.ei.xstat
#define p_realtimer	ps.ii.realtimer
#define p_itimerlock	ps.ii.itimerlock
#define p_itimer_tick	ps.ii.interval_tick
#define p_itimer_unit	ps.ii.interval_unit

/* p_itimerunit */
#define FAST_TICK	1
#define SLOW_TICK	0

/* p_segtype */
#define PSEG_NULL	0		/* no segment table */
#define PSEG_TABLE	1		/* seg table is a full table */
#define PSEG_LIST	2		/* seg table is a list */

/* stat codes */

#define	SSLEEP	1		/* Awaiting an event.		*/
#define	SRUN	2		/* Running.			*/
#define	SZOMB	3		/* Process terminated but not	*/
				/* waited for.			*/
#define	SSTOP	4		/* Process stopped by signal	*/
				/* since it is being traced by	*/
				/* its parent.			*/
#define	SIDL	5		/* Intermediate state in	*/
				/* process creation.		*/
#define SXBRK	7		/* process being xswapped       */

/* flag codes */

#define	SSYS	0x0001		/* System (resident) process.	*/
#define	STRC	0x0002		/* Process is being traced.	*/
#define	SWTED	0x0004		/* Stopped process has been	*/
				/* given to parent by wait	*/
				/* system call.  Don't return	*/
				/* this process to parent again	*/
				/* until it runs first.		*/
#define SNWAKE	0x0008		/* Process cannot wakeup by	*/
				/* a signal.			*/
#define SLOAD	0x00000010	/* In core                      */
#define SULOAD	0x00000020	/* ublock in core (was SLOCK)	*/
#define	SATMSIM	0x00000040	/* Look for atomic operation simulation */
#define SPOLL	0x00000080	/* Process in stream poll	*/
#define SSEL	SPOLL		/* 4.3 select(1) hack		*/
#define SPRSTOP	0x00000100	/* process is being stopped via /proc */
#define SPROCTR	0x00000200	/* sysent/exit/fault/signal tracing via /proc */
#define	SPARSYS	0x00000400	/* tracing sys calls via par */
#define SJSTOP	0x00000800	/* process wants to job control stop */
#define SPROPEN	0x00001000	/* process is open via /proc	*/
#define SUWANT	0x00002000	/* want ublock in core		*/
#define	SNOCTTY	0x00004000	/* no control terminal */
#define	SOWEUPC	0x00008000	/* owe user an addupc call (profiling tick) */
#define SWSYS	0x00010000	/* process got watchpoint in system */
#define	SULTPG	0x00020000	/* process can claim last page of memory */
#define SEXIT	0x00040000	/* process is exiting		*/
#define SBSDSTOP 0x00080000	/* proces did BSDsetpgrp so send CLD */
#define SSHFSYN 0x00100000	/* process needs to sync open file descr */
#define SSHDSYN 0x00200000	/* process needs to sync current dir */
#define SSHUSYN 0x00400000	/* process needs to sync uids	*/
#define SSHLSYN 0x00800000	/* process needs to sync ulimit	*/
#define SSHMSYN 0x01000000	/* process needs to sync umask	*/
#define SBLOCK	0x02000000	/* process may need to block itself */
#define SPROF	0x04000000	/* process is profiling */
#define SSTEP	0x08000000	/* process is single stepping */
#define SNOCLDSTOP 0x10000000	/* SIGCHLD not sent when child stops */
#define SEXECED	0x20000000	/* process successfully exec'ed (for setpgid) */
#define SRECALC	0x40000000	/* recalc user pri at setrq time */
#define SGRAPH  0x80000000	/* member of share group just given graphics */

#define PTRACED(p)	((p)->p_flag&(STRC|SPROCTR|SPROPEN))
#define PSYNCFLAGS	(SSHFSYN|SSHDSYN|SSHUSYN|SSHLSYN|SSHMSYN)

/* values for p_flag2 */
#define SNOSWAP	0x00000001	/* communicate no swap condition vhand->proc */
#define SCEXIT	0x00000002	/* this process will die when parent exits */
#define SBBLST	0x00000004	/* its been blasted by sched deadlock scan */
#define SCOLL	0x00000008	/* collision while select/polling */
#define SWSRCH	0x00000010	/* process is scanning in wait */
#define	SFIXADE	0x00000020	/* fixup unaligned address errors */
#define SPARTRC 0x00000040	/* par is tracing this */
#define SONCESTOPPED 0x00000080	/* won't run in user space again */

/* flags for p_fp */
#define	P_FP_SIGINTR1	1
#define	P_FP_SIGINTR2	2

/* Flags for newproc() */
#define NP_FAILOK	0x1	/* don't panic if cannot create process */
#define NP_NOLAST	0x2	/* don't use last process slot */
#define	NP_SYSPROC	0x4	/* system (resident) process */

/* Reasons for stopping (values of p_whystop) */
#define	REQUESTED	1
#define	SIGNALLED	2
#define	SYSENTRY	3
#define	SYSEXIT		4
#define FAULTED		5
#define JOBCONTROL	6

/* Reasons for stopping (p_whatstop) */
#define FAULTEDWATCH	1	/* whystop == FAULTED */
#define FAULTEDKWATCH	2	/* whystop == FAULTED */
#define FAULTEDPAGEIN	3	/* whystop == FAULTED */
#define FAULTEDSTACK	4	/* whystop == FAULTED */

/* Reasons for calling issig() */
#define	FORREAL		0	/* Usual side-effects */
#define	JUSTLOOKING	1	/* Don't stop the process */

#define ISSIG(p, l, w)	((u.u_cursig || (p)->p_sig) && issig(l, w))

#ifdef _KERNEL

extern struct proc	proc[];		/* the proc table itself */

/* fault trace variables */
extern proc_t		*trproc;	/* process being traced */
extern int		trflags;	/* trace flags */

#define	isresident(p)	(p)->p_resident

extern proc_t *procnew(int);	/* allocate a new proc entry */
extern int procscan();
extern int proccscan();
extern proc_t *procfind(pid_t);
extern void procunlock(proc_t *);
extern proc_t *procalloc(void);
extern void procfree(proc_t *);
extern void procdeact(proc_t *);
extern void procchain(proc_t *, proc_t *);
extern void procunchain(proc_t *);
extern int proc1chain(proc_t *);
extern int inferior(proc_t *);
extern int ancestor(proc_t *);
extern void prochashtables(void);
extern void recalcrss(proc_t *);

extern void signal(int, int);
extern void psignal(proc_t *, int);
extern int issig(int, int);
extern int isfatalsig(proc_t *);
extern ulong hold_nonfatalsig(proc_t *);
extern void release_nonfatalsig(proc_t *, ulong);
extern int psig(int);
extern int checkstop(void);
extern void stop(proc_t *, ushort, ushort);
extern int fsig(proc_t *, int);
extern void freeproc(proc_t *, int);
extern void wait4(int *, int, struct rusage *, pid_t);
extern void setprun(proc_t *, int);
extern void setrun(proc_t *);
extern void setrq(proc_t *);
extern int wantdisp(proc_t *);
extern char chgrtpri(proc_t *, char);
extern int chgpri(int, int);
extern int resident(proc_t *);
extern void nonresident(proc_t *);

/* context switch routines */
extern void qswtch(int);
extern void pswtch(int);
extern void swtch(int);
extern void presume(proc_t *, int);
extern int save(void);
extern void resume(proc_t *, int, unsigned char);
extern void resumeidle(int);
extern proc_t *disp(void);

#endif /* _KERNEL */

/*
 * shared group descriptor
 * Each process that sproc()'s has one of these
 *
 * Locks/Semaphores:
 *
 *	s_listlock - also for modifying/accessing the list s_plink: use shared
 *		shared lock for long term access, or use this for short term
 */

typedef struct shaddr_s {
	/* the following are all to handle VM */
	struct pregion	*s_region;	/* processes' shared pregions */
	mrlock_t	s_aspacelock;	/* address space lock */
	/* generic fields for handling share groups */
	struct proc	*s_plink;	/* link to processes in share group */
	ushort		s_refcnt;	/* # process in list */
	ushort		s_flag;		/* flags */
	lock_t		s_listlock;	/* protects s_plink */
	/* semaphore for single threading open file updating */
	sema_t		s_fupdsema;	/* wait for opening file */
	struct file	**s_ofile;	/* copy of open file descriptors */
	char		*s_pofile;	/* copy of open file flags */
	struct inode	*s_cdir;	/* current directory */
	struct inode	*s_rdir;	/* root directory */
	/* lock for updating misc things that don't need a semaphore */
	lock_t		s_rupdlock;	/* update lock */
	/* hold values for other sharing options */
	char		s_exitsig;	/* signal to send on member death */
	unchar		s_sched;	/* scheduling mode of share gp */
	short		s_cmask;	/* mask for file creation */
	daddr_t		s_limit;	/* maximum write address */
	uid_t		s_uid;		/* effective user id */
	gid_t		s_gid;		/* effective group id */
	gid_t		s_groups[NGROUPS_UMAX];	/* BSD multiple groups */
	/* semaphore for single threading page locking */
	sema_t		s_pinsema;	/* single thread mpin */
	int		s_nlockpg;	/* # locked pages		*/
	/* share group scheduling information (protected by s_rupdlock) */
	struct proc	*s_master;	/* proc that runs when in master mode */
	ushort		s_runnable;	/* # of gp members running or on runq */
	/* fields for graphics share groups */
	int		s_gfxruncount;	/* num of gfx gp members running */
	int		s_gfxreaders;	/* num of readers reading */
	sema_t		s_gfxwait;	/* waiting readers or writers */
	lock_t		s_gfxlock;	/* lock for above fields */
} shaddr_t;

/* types for grabbing shared block pregion lock */
#define SH_ACC		MR_ACCESS
#define SH_UPD		MR_UPDATE

#ifdef _KERNEL
/* lock and unlock address space */
#define shdlock(p, t)		mrlock((p)->p_aspacelock, t, PZERO);
#define shdunlock(p)		mrunlock((p)->p_aspacelock);

#define cshdlock(p, t)		cmrlock((p)->p_aspacelock, t)
#define IS_LCKUPD(p)		ismrlocked((p)->p_aspacelock, MR_UPDATE)
#define IS_LCKACC(p)		ismrlocked((p)->p_aspacelock, MR_ACCESS)

/* lock and unlock shared file descriptor updating */
#define shdflock()		{ if (u.u_procp->p_shaddr) sharedf_lock(); }
#define shdfunlock()		{ if (u.u_procp->p_shaddr) sharedf_unlock(); }
#define IS_FLCK(p)		(valusema(&(p)->p_shaddr->s_fupdsema) <= 0)

/* lock/unlock misc update lock */
#define shdrlock()		{ if (u.u_procp->p_shaddr) \
			spsemahi(u.u_procp->p_shaddr->s_rupdlock); }
#define shdrunlock()		{ if (u.u_procp->p_shaddr) \
			svsemax(u.u_procp->p_shaddr->s_rupdlock); }

/* lock/unlock mpin semaphore */
#define shdpinlock()		{ if (u.u_procp->p_shaddr) \
			psema(&u.u_procp->p_shaddr->s_pinsema, PZERO); }
#define shdpinunlock()		{ if (u.u_procp->p_shaddr) \
			vsema(&u.u_procp->p_shaddr->s_pinsema); }

extern int allocshaddr(proc_t *);
extern void attachshaddr(proc_t *, proc_t *);
extern void detachshaddr(struct user *);
extern void sharedf_lock(void);
extern void sharedf_unlock(void);
#endif /* _KERNEL */

typedef struct pgroup_s {
        pid_t   	pg_pgrp;        /* name of process group leader */
	uint		pg_flag;	/* process group status */
	struct proc	*pg_chain;	/* process group chain */
	uint		pg_refcnt;	/* count of process group */
	struct pgroup_s *pg_next;	/* hash chain */
} pgroup_t;

#define	PGORPHANED	0x0001		/* this process detached from ctty */

typedef struct session_s {
	pid_t		se_sid;		/* session ID */
	uint		se_refcnt;	/* reference count. it is protected
					   by semaphore pglobal */
	pid_t		*se_ttyp;	/* foreground process group ID */
	pid_t		*se_ttys;	/* foreground process session ID */
	struct inode	*se_ip;		/* tty's inode structure */
	uint		se_ttycnt;	/* count for /dev/tty */
	lock_t		se_lock;	/* lock for se_ttycnt */
	sema_t		se_sema;	/* process waiting for /dev/tty */
	uint		se_flag;	/* flags for session */
} session_t;

#define	SETTYCLOSE	0x0001		/* ctty detached from session*/

#ifdef _KERNEL

enum jobcontrol {
        JCTLREAD,	                 /* read data on a ctty */
        JCTLWRITE,	                 /* write data to a ctty */
        JCTLGST		                 /* get ctty parameters */
};

/*
 * Add a process to an existing session
 */
#define JOIN_SESS(p, sp)	{     (p)->p_session = (sp); \
					(sp)->se_refcnt++; \
				}

/*
 * LEAVE_SESS:
 * update session reference count.  if the reference count is zero, 
 * deallocate the session.
 */
 
#define LEAVE_SESS(p) 	{	ASSERT(p->p_session->se_refcnt >0); \
			 	if (--(p->p_session->se_refcnt) == 0) { \
					ASSERT(p->p_session != &session0); \
					freesema(&p->p_session->se_sema); \
					freesplock(p->p_session->se_lock); \
					kern_free((caddr_t)p->p_session); \
				} \
				p->p_session = NULL; \
			}

extern sema_t pglobal;			/* global process group semaphore */
extern lock_t pgchainlck;		/* global process chain lock */


extern int pgrpscan();
extern void detachildren(proc_t *);
extern void joinpg(proc_t *, pid_t, int);
extern void leavepg(proc_t *, int);
extern pgroup_t *findpg(pid_t);
extern int pgrpverify(proc_t *, pid_t);
extern int memberspg(proc_t *, pid_t);
extern session_t session0;
extern session_t *alloc_session(void);
extern void alloctty(proc_t *, struct inode *, pid_t *, pid_t *);
extern void freectty(session_t *);
extern dev_t cttydev(proc_t *);
extern int findsession(proc_t *, struct inode *, int);

/* process tlb routines */
extern void new_tlbpid(proc_t *, int);
extern void newptes(proc_t *, caddr_t, uint);
extern int tlbsync(shaddr_t *, int);
extern void tlbdirty(proc_t *);
extern void tlbclean(proc_t *, uint);
extern void set_tlbpids(proc_t *, unsigned char);

#endif /* _KERNEL */
#endif /* __SYS_PROC_H__ */
